from __future__ import annotations

import functools
import io
import re
import shlex
import sys
import tokenize
from typing import Any, Callable, Generic, TypeVar

from ssort._exceptions import UnknownEncodingError

if sys.version_info < (3, 9):
    memoize = functools.lru_cache(maxsize=None)
else:
    memoize = functools.cache


def sort_key_from_iter(values):
    index = {statement: index for index, statement in enumerate(values)}
    key = lambda value: index[value]
    return key


_T = TypeVar("_T")


class _SingleDispatch(Generic[_T]):
    """A more performant implementation of functools.singledispatch."""

    def __init__(self, function: Callable[..., _T]) -> None:
        functools.update_wrapper(self, function)
        self._function: Callable[..., _T] = function
        self._functions: dict[type[Any], Callable[..., _T]] = {}

    def register(
        self, cls: type[Any]
    ) -> Callable[[Callable[..., _T]], Callable[..., _T]]:
        def decorator(function: Callable[..., _T]) -> Callable[..., _T]:
            self._functions[cls] = function
            return function

        return decorator

    def __call__(self, arg: Any, *args: Any) -> _T:
        return self._functions.get(type(arg), self._function)(arg, *args)


single_dispatch = _SingleDispatch


def cached_method(function: Callable[[Any], _T]) -> Callable[[Any], _T]:
    cached_attribute_name = f"_{function.__name__}_cache"

    @functools.wraps(function)
    def wrapper(self) -> _T:
        try:
            return getattr(self, cached_attribute_name)
        except AttributeError:
            value = function(self)
            setattr(self, cached_attribute_name, value)
            return value

    return wrapper


def escape_path(path):
    """
    Takes a `pathlib.Path` object and returns a string representation that can
    be safely copied into the system shell.
    """
    if sys.platform == "win32":
        # TODO
        return str(path)
    else:
        return shlex.quote(str(path))


def detect_encoding(bytestring):
    """
    Detect the encoding of a python source file based on "coding" comments, as
    defined in [PEP 263](https://www.python.org/dev/peps/pep-0263/).
    """
    try:
        encoding, _ = tokenize.detect_encoding(io.BytesIO(bytestring).readline)
    except SyntaxError as exc:
        raise UnknownEncodingError(
            exc.msg, encoding=re.match("unknown encoding: (.*)", exc.msg)[1]
        ) from exc
    return encoding


_NEWLINE_RE = re.compile("(\r\n)|(\r)|(\n)")


def detect_newline(text):
    """
    Detects the newline character used in a source file based on the first
    occurence of '\\n', '\\r' or '\\r\\n'.
    """
    match = re.search(_NEWLINE_RE, text)
    if match is None:
        return "\n"
    return match[0]


def normalize_newlines(text):
    """
    Replaces all occurrences of '\r' and '\\r\\n' with \n.
    """
    return re.sub(_NEWLINE_RE, "\n", text)



from __future__ import annotations

import ast
from typing import Iterable

from ssort._bindings import get_bindings
from ssort._method_requirements import get_method_requirements
from ssort._requirements import Requirement, get_requirements
from ssort._utils import cached_method


class Statement:
    def __init__(
        self, *, text: str, node: ast.AST, start_row: int, start_col: int
    ) -> None:
        self.text = text
        self.node = node
        self.start_row = start_row
        self.start_col = start_col

    @cached_method
    def text_padded(self) -> str:
        """
        Return the statement text padded with leading whitespace so that
        coordinates in the ast match up with coordinates in the text.
        """
        return ("\n" * self.start_row) + (" " * self.start_col) + self.text

    @cached_method
    def requirements(self) -> Iterable[Requirement]:
        """
        Returns an iterable yielding Requirement objects describing the
        bindings that this statement references.
        """
        return tuple(get_requirements(self.node))

    @cached_method
    def method_requirements(self) -> Iterable[str]:
        """
        Returns an iterable yielding the names of attributes of the `self`
        parameter that this statement depends on.
        """
        return tuple(get_method_requirements(self.node))

    @cached_method
    def bindings(self) -> Iterable[str]:
        """
        Returns an iterable yielding the names bound by this statement.
        """
        return tuple(get_bindings(self.node))

    def __repr__(self) -> str:
        return f"<Statement text={self.text!r}>"



from __future__ import annotations

import ast
import sys
from typing import Iterable

from ssort._ast import iter_child_nodes
from ssort._utils import single_dispatch


@single_dispatch
def get_bindings(node: ast.AST) -> Iterable[str]:
    for child in iter_child_nodes(node):
        yield from get_bindings(child)


@get_bindings.register(ast.FunctionDef)
@get_bindings.register(ast.AsyncFunctionDef)
def _get_bindings_for_function_def(
    node: ast.FunctionDef | ast.AsyncFunctionDef,
) -> Iterable[str]:
    for decorator in node.decorator_list:
        yield from get_bindings(decorator)
    yield node.name
    yield from get_bindings(node.args)
    if node.returns is not None:
        yield from get_bindings(node.returns)


@get_bindings.register(ast.ClassDef)
def _get_bindings_for_class_def(node: ast.ClassDef) -> Iterable[str]:
    for decorator in node.decorator_list:
        yield from get_bindings(decorator)
    for base in node.bases:
        yield from get_bindings(base)
    for keyword in node.keywords:
        yield from get_bindings(keyword.value)
    yield node.name


@get_bindings.register(ast.Import)
def _get_bindings_for_import(node: ast.Import) -> Iterable[str]:
    for name in node.names:
        if name.asname:
            yield name.asname
        else:
            root, *rest = name.name.split(".", 1)
            yield root


@get_bindings.register(ast.ImportFrom)
def _get_bindings_for_import_from(node: ast.ImportFrom) -> Iterable[str]:
    for name in node.names:
        yield name.asname if name.asname else name.name


@get_bindings.register(ast.Global)
def _get_bindings_for_global(node: ast.Global) -> Iterable[str]:
    yield from node.names


@get_bindings.register(ast.Nonlocal)
def _get_bindings_for_nonlocal(node: ast.Nonlocal) -> Iterable[str]:
    yield from node.names


@get_bindings.register(ast.Lambda)
def _get_bindings_for_lambda(node: ast.Lambda) -> Iterable[str]:
    yield from get_bindings(node.args)


@get_bindings.register(ast.Name)
def _get_bindings_for_name(node: ast.Name) -> Iterable[str]:
    if isinstance(node.ctx, ast.Store):
        yield node.id


@get_bindings.register(ast.comprehension)
def _get_bindings_for_comprehension(node: ast.comprehension) -> Iterable[str]:
    # Ignore target, it can never produce bindings
    yield from get_bindings(node.iter)
    for condition in node.ifs:
        yield from get_bindings(condition)


@get_bindings.register(ast.ExceptHandler)
def _get_bindings_for_except_handler(node: ast.ExceptHandler) -> Iterable[str]:
    if node.type:
        yield from get_bindings(node.type)
    if node.name:
        yield node.name
    for statement in node.body:
        yield from get_bindings(statement)


if sys.version_info >= (3, 10):

    @get_bindings.register(ast.MatchStar)
    def _get_bindings_for_match_star(node: ast.MatchStar) -> Iterable[str]:
        if node.name is not None:
            yield node.name

    @get_bindings.register(ast.MatchMapping)
    def _get_bindings_for_match_mapping(
        node: ast.MatchMapping,
    ) -> Iterable[str]:
        for key in node.keys:
            yield from get_bindings(key)
        for pattern in node.patterns:
            yield from get_bindings(pattern)
        if node.rest is not None:
            yield node.rest

    @get_bindings.register(ast.MatchAs)
    def _get_bindings_for_match_as(node: ast.MatchAs) -> Iterable[str]:
        if node.pattern is not None:
            yield from get_bindings(node.pattern)
        if node.name is not None:
            yield node.name



from ssort._builtins import MODULE_BUILTINS
from ssort._graphs import Graph


def module_statements_graph(statements, *, on_unresolved, on_wildcard_import):
    """
    Constructs a graph of the interdependencies in a list of module level
    statements.

    :param statements:
        An ordered list of opaque `Statement` objects from which to construct
        the graph.

    :param on_unresolved:
        An callback that should be invoked for each unresolved dependency.  Can
        safely raise any arbitrary exception to abort constructing the graph.
        If no exception is raised, the graph returned by this function will not
        contain a link for the missing requirement.
    :param on_wildcard_import:
        A callback that should be invoked if ssort detects a `*` import.  If no
        exception is raised, all dangling references will be pointed back to the
        last `*` import.

    :returns:
        A `Graph` mapping from statements to the set of statements that they
        depend on.
    """
    # A dictionary mapping from names to the statements which bind them.
    scope = {}

    all_requirements = []
    resolved = {}

    for statement in statements:
        for requirement in statement.requirements():
            all_requirements.append(requirement)

            # TODO error if requirement is not deferred.
            if requirement.name in scope:
                resolved[requirement] = scope[requirement.name]
                continue

            if requirement.name in MODULE_BUILTINS:
                resolved[requirement] = None
                continue

        for name in statement.bindings():
            if name == "*":
                on_wildcard_import(
                    lineno=statement.node.lineno,
                    col_offset=statement.node.col_offset,
                )

            scope[name] = statement

    # Patch up dependencies that couldn't be resolved immediately.
    for requirement in all_requirements:
        if requirement in resolved:
            continue

        if requirement.name not in scope:
            continue

        resolved[requirement] = scope[requirement.name]

    if "*" in scope:
        for requirement in all_requirements:
            if requirement in resolved:
                continue

            resolved[requirement] = scope["*"]

    else:
        unresolved = [
            requirement
            for requirement in all_requirements
            if requirement not in resolved
        ]

        for requirement in unresolved:
            on_unresolved(
                f"could not resolve {requirement.name!r}",
                name=requirement.name,
                lineno=requirement.lineno,
                col_offset=requirement.col_offset,
            )

        if unresolved:
            # Not safe to attempt to re-order in the event of unresolved
            # dependencies.  A typo could cause a statement to be moved ahead of
            # something that it should depend on.
            return None

    graph = Graph()
    for statement in statements:
        graph.add_node(statement)

    for statement in statements:
        for requirement in statement.requirements():
            if resolved[requirement] is not None:
                graph.add_dependency(statement, resolved[requirement])

    # Add links between statements that overwrite the same binding to make sure
    # that bindings are always applied in the same order.
    scope = {}
    for statement in statements:
        for name in statement.bindings():
            if name in scope:
                graph.add_dependency(statement, scope[name])
            scope[name] = statement

    return graph


def class_statements_initialisation_graph(statements):
    """
    Constructs a graph of the hard dependencies within a list of class level
    statements.  These are dependencies that absolutely cannot be reordered
    without changing the semantics of the script.

    At inititialisation, class level statements can see earlier bindings in the
    class body.

    Note that this isn't a proper scope: variables from the containing scope are
    not shadowed until after a binding is actually made and, obviously from the
    need for a `self` argument, bindings are not captured.

    The following example demonstrates some of this behaviour:

    .. code:: python

        >>> r = 1
        >>> class A:
        ...     a = r
        ...     r = 2
        ...     b = r
        >>> r
        1
        >>> A.a
        1
        >>> A.b
        2

    If a name cannot be resolved at the class level we assume that it is
    resolved at the module level and don't emit any warning.

    :param statements:
        An ordered list of opaque `Statement` objects from which to construct
        the graph.

    :returns:
        A `Graph` mapping from statements to the set of statements that they
        depend on.
    """
    scope = {}

    graph = Graph()

    for statement in statements:
        graph.add_node(statement)

        for requirement in statement.requirements():
            if requirement.deferred:
                continue

            if requirement.name not in scope:
                continue

            graph.add_dependency(statement, scope[requirement.name])

        for name in statement.bindings():
            scope[name] = statement

    return graph


def class_statements_runtime_graph(statements, *, ignore_public):
    """
    Attempts to construct a graph of the soft, runtime dependencies between the
    methods of a class.

    The graph is inferred by looking for attribute access on the `self` argument
    of methods.

    :param statements:
        An ordered list of opaque `Statement` objects from which to construct
        the graph.
    :param ignore_public:
        If true, the graph will only include references to private attributes,
        i.e attributes prefixed by `_`.  This leaves the sorting of public
        methods, which make up the interface of the class, to the programmer.

    :returns:
        A `Graph` mapping from statements to the set of statements that they
        depend on.
    """
    scope = {}

    graph = Graph()

    for statement in statements:
        graph.add_node(statement)

        for name in statement.bindings():
            scope[name] = statement

    for statement in statements:
        for name in statement.method_requirements():
            if ignore_public and not name.startswith("_"):
                continue
            if name not in scope:
                continue
            graph.add_dependency(statement, scope[name])

    return graph



"""
The python source code statement sorter.
"""
from ssort._exceptions import (
    DecodingError,
    ParseError,
    ResolutionError,
    UnknownEncodingError,
    WildcardImportError,
)
from ssort._ssort import ssort

# Let linting tools know that we do mean to re-export exception classes.
assert DecodingError is not None
assert ParseError is not None
assert ResolutionError is not None
assert UnknownEncodingError is not None
assert WildcardImportError is not None

try:
    from ssort._version import VERSION as __version__  # type: ignore
except ImportError:
    __version__ = "0.0.1+dev"

__all__ = ["ssort"]



class UnknownEncodingError(Exception):
    def __init__(self, msg, *, encoding):
        super().__init__(msg)
        self.encoding = encoding


class DecodingError(Exception):
    pass


class ParseError(Exception):
    def __init__(self, msg, *, lineno, col_offset):
        super().__init__(msg)
        self.lineno = lineno
        self.col_offset = col_offset


class ResolutionError(Exception):
    def __init__(self, msg, *, name, lineno, col_offset):
        super().__init__(msg)
        self.name = name
        self.lineno = lineno
        self.col_offset = col_offset


class WildcardImportError(Exception):
    def __init__(self, msg, *, lineno, col_offset):
        super().__init__(msg)
        self.lineno = lineno
        self.col_offset = col_offset



import ast
import re
import sys

from ssort._dependencies import (
    class_statements_initialisation_graph,
    class_statements_runtime_graph,
    module_statements_graph,
)
from ssort._exceptions import (
    DecodingError,
    ParseError,
    ResolutionError,
    UnknownEncodingError,
    WildcardImportError,
)
from ssort._graphs import (
    is_topologically_sorted,
    replace_cycles,
    topological_sort,
)
from ssort._parsing import parse, split_class
from ssort._utils import (
    detect_encoding,
    detect_newline,
    normalize_newlines,
    sort_key_from_iter,
)

SPECIAL_PROPERTIES = [
    "__doc__",
    "__slots__",
]

LIFECYCLE_OPERATIONS = [
    # Lifecycle.
    "__new__",
    "__init__",
    "__del__",
    # Copying.
    "__copy__",
    "__deepcopy__",
    # Metaclasses.
    # TODO "__prepare__", ?
    "__init_subclass__",
    "__instancecheck__",
    "__subclasscheck__",
    # Generics.
    "__class_getitem__",
    # Descriptors.
    "__get__",
    "__set__",
    "__delete__",
    "__set_name__",
]

REGULAR_OPERATIONS = [
    # Callables.
    "__call__",
    # Attribute Access.
    "__getattr__",
    "__getattribute__",
    "__setattr__",
    "__delattr__",
    "__dir__",
    # Container Operations.
    "__getitem__",
    "__setitem__",
    "__delitem__",
    "__missing__",
    "__iter__",
    "__reversed__",
    "__contains__",
    "__len__",
    "__length_hint__",
    # Binary Operators.
    "__add__",
    "__radd__",
    "__iadd__",
    "__sub__",
    "__rsub__",
    "__isub__",
    "__mul__",
    "__rmul__",
    "__imul__",
    "__matmul__",
    "__rmatmul__",
    "__imatmul__",
    "__truediv__",
    "__rtruediv__",
    "__itruediv__",
    "__floordiv__",
    "__rfloordiv__",
    "__ifloordiv__",
    "__mod__",
    "__rmod__",
    "__imod__",
    "__divmod__",
    "__rdivmod__",
    "__pow__",
    "__rpow__",
    "__ipow__",
    "__lshift__",
    "__rlshift__",
    "__ilshift__",
    "__rshift__",
    "__rrshift__",
    "__irshift__",
    "__and__",
    "__rand__",
    "__iand__",
    "__xor__",
    "__rxor__",
    "__ixor__",
    "__or__",
    "__ror__",
    "__ior__",
    # Unary operators.
    "__neg__",
    "__pos__",
    "__abs__",
    "__invert__",
    # Rich comparison operators.
    "__lt__",
    "__le__",
    "__eq__",
    "__ne__",
    "__gt__",
    "__ge__",
    "__hash__",
    # Numeric conversions
    "__bool__",
    "__complex__",
    "__int__",
    "__float__",
    "__index__",
    "__round__",
    "__trunc__",
    "__floor__",
    "__ceil__",
    # Context managers.
    "__enter__",
    "__exit__",
    # Async tasks.
    "__await__",
    # Async iterators.
    "__aiter__",
    "__anext__",
    # Async context managers.
    "__aenter__",
    "__aexit__",
    # Pickling.
    "__getnewargs_ex__",
    "__reduce__",
    "__getstate__",
    "__setstate__",
    # Formatting.
    "__repr__",
    "__str__",
    "__bytes__",
    "__format__",
]


def _partition(values, predicate):
    passed = []
    failed = []

    for value in values:
        if predicate(value):
            passed.append(value)
        else:
            failed.append(value)
    return passed, failed


def _is_string(statement):
    expr_node = statement.node
    if not isinstance(expr_node, ast.Expr):
        return False

    node = expr_node.value
    if not isinstance(node, ast.Constant):
        return False

    return isinstance(node.value, str)


def _is_special_property(statement):
    return any(
        binding in SPECIAL_PROPERTIES for binding in statement.bindings()
    )


def _is_lifecycle_operation(statement):
    return any(
        binding in LIFECYCLE_OPERATIONS for binding in statement.bindings()
    )


def _is_regular_operation(statement):
    return any(
        binding in REGULAR_OPERATIONS for binding in statement.bindings()
    )


def _is_property(statement):
    return isinstance(
        statement.node, (ast.Assign, ast.AnnAssign, ast.AugAssign)
    )


def _is_class(statement):
    return isinstance(statement.node, ast.ClassDef)


def _statement_binding_sort_key(binding_key):
    def _safe_binding_key(binding):
        try:
            return binding_key(binding)
        except KeyError:
            return sys.maxsize

    def _key(statement):
        bindings = statement.bindings()
        return min(_safe_binding_key(binding) for binding in bindings)

    return _key


def _statement_text_sorted_class(statement):
    head_text, statements = split_class(statement)

    # Take a snapshot of any hard dependencies between statements so that we can
    # restore them later.
    initialisation_graph = class_statements_initialisation_graph(statements)

    # === Split up the statements into high level groups =======================
    if _is_string(statements[0]):
        docstrings, statements = statements[:1], statements[1:]
    else:
        docstrings = []

    special_properties, statements = _partition(
        statements, _is_special_property
    )

    lifecycle_operations, statements = _partition(
        statements, _is_lifecycle_operation
    )

    regular_operations, statements = _partition(
        statements, _is_regular_operation
    )

    inner_classes, statements = _partition(statements, _is_class)

    properties, statements = _partition(statements, _is_property)

    methods, statements = statements, []

    sorted_statements = []

    # === Join groups back together in the correct order =======================
    sorted_statements += docstrings

    # Special properties (in hard-coded order).
    sorted_statements += sorted(
        special_properties,
        key=_statement_binding_sort_key(
            sort_key_from_iter(SPECIAL_PROPERTIES)
        ),
    )

    # Inner classes (in original order).
    sorted_statements += inner_classes

    # Regular properties (in original order).
    sorted_statements += properties

    # Special lifecycle methods (in hard-coded order).
    sorted_statements += sorted(
        lifecycle_operations,
        key=_statement_binding_sort_key(
            sort_key_from_iter(LIFECYCLE_OPERATIONS)
        ),
    )

    # Regular methods.
    sorted_statements += methods

    # Special operations (in hard-coded order).
    sorted_statements += sorted(
        regular_operations,
        key=_statement_binding_sort_key(
            sort_key_from_iter(REGULAR_OPERATIONS)
        ),
    )

    # === Re-sort based on dependencies between statements =====================

    # Fix any hard dependencies.
    sorted_statements = topological_sort(
        sorted_statements, graph=initialisation_graph
    )

    # Attempt to resolve soft dependencies on private attributes, but with hard
    # dependencies taking priority, and always preserving the original order
    # where there are cycles.
    runtime_graph = class_statements_runtime_graph(
        sorted_statements, ignore_public=True
    )
    runtime_graph.update(initialisation_graph)
    replace_cycles(runtime_graph, key=sort_key_from_iter(sorted_statements))

    sorted_statements = topological_sort(
        sorted_statements, graph=runtime_graph
    )

    return (
        head_text
        + "\n"
        + "\n".join(
            statement_text_sorted(body_statement)
            for body_statement in sorted_statements
        )
    )


def statement_text_sorted(statement):
    node = statement.node
    if isinstance(node, ast.ClassDef):
        return _statement_text_sorted_class(statement)
    return statement.text


def _on_unknown_encoding_ignore(message, **kwargs):
    pass


def _on_unknown_encoding_raise(message, *, encoding, **kwargs):
    raise UnknownEncodingError(message, encoding=encoding)


def _interpret_on_unknown_encoding_action(on_unknown_encoding):
    if on_unknown_encoding == "ignore":
        return _on_unknown_encoding_ignore

    if on_unknown_encoding == "raise":
        return _on_unknown_encoding_raise

    return on_unknown_encoding


def _on_decoding_error_ignore(message, **kwargs):
    pass


def _on_decoding_error_raise(message, **kwargs):
    raise DecodingError(message)


def _interpret_on_decoding_error_action(on_decoding_error):
    if on_decoding_error == "ignore":
        return _on_decoding_error_ignore

    if on_decoding_error == "raise":
        return _on_decoding_error_raise

    return on_decoding_error


def _on_parse_error_ignore(message, **kwargs):
    pass


def _on_parse_error_raise(message, *, lineno, col_offset, **kwargs):
    raise ParseError(message, lineno=lineno, col_offset=col_offset)


def _interpret_on_parse_error_action(on_parse_error):
    if on_parse_error == "ignore":
        return _on_parse_error_ignore

    if on_parse_error == "raise":
        return _on_parse_error_raise

    return on_parse_error


def _on_unresolved_ignore(message, *, name, lineno, col_offset, **kwargs):
    pass


def _on_unresolved_raise(message, *, name, lineno, col_offset, **kwargs):
    raise ResolutionError(
        message,
        name=name,
        lineno=lineno,
        col_offset=col_offset,
    )


def _interpret_on_unresolved_action(on_unresolved):
    if on_unresolved == "ignore":
        return _on_unresolved_ignore

    if on_unresolved == "raise":
        return _on_unresolved_raise

    return on_unresolved


def _on_wildcard_import_ignore(**kwargs):
    pass


def _on_wildcard_import_raise(*, lineno, col_offset, **kwargs):
    raise WildcardImportError(
        "can't reliably determine dependencies on * import",
        lineno=lineno,
        col_offset=col_offset,
    )


def _interpret_on_wildcard_import_action(on_wildcard_import):
    if on_wildcard_import == "ignore":
        return _on_wildcard_import_ignore

    if on_wildcard_import == "raise":
        return _on_wildcard_import_raise

    return on_wildcard_import


def ssort(
    text,
    *,
    filename="<unknown>",
    on_unknown_encoding_error="raise",
    on_decoding_error="raise",
    on_parse_error="raise",
    on_unresolved="raise",
    on_wildcard_import="raise",
):
    on_unknown_encoding_error = _interpret_on_unknown_encoding_action(
        on_unknown_encoding_error
    )
    on_decoding_error = _interpret_on_decoding_error_action(on_decoding_error)
    on_parse_error = _interpret_on_parse_error_action(on_parse_error)
    on_unresolved = _interpret_on_unresolved_action(on_unresolved)
    on_wildcard_import = _interpret_on_wildcard_import_action(
        on_wildcard_import
    )

    try:
        encoding = None
        if isinstance(text, bytes):
            encoding = detect_encoding(text)
            text = text.decode(encoding)
    except UnknownEncodingError as exc:
        on_unknown_encoding_error(str(exc), encoding=exc.encoding)
        return text

    except UnicodeDecodeError as exc:
        on_decoding_error(str(exc))
        return text

    newline = detect_newline(text)
    text = normalize_newlines(text)

    try:
        statements = list(parse(text, filename=filename))
    except ParseError as exc:
        on_parse_error(str(exc), lineno=exc.lineno, col_offset=exc.col_offset)
        return text

    if not statements:
        return text

    graph = module_statements_graph(
        statements,
        on_unresolved=on_unresolved,
        on_wildcard_import=on_wildcard_import,
    )
    if graph is None:
        return text

    replace_cycles(graph, key=sort_key_from_iter(statements))

    sorted_statements = topological_sort(statements, graph=graph)

    assert is_topologically_sorted(sorted_statements, graph=graph)

    output = "\n".join(
        statement_text_sorted(statement) for statement in sorted_statements
    )
    if output:
        output += "\n"

    if newline != "\n":
        output = re.sub("\n", newline, output)
    if encoding is not None:
        output = output.encode(encoding)
    return output



from __future__ import annotations

import ast
from typing import Iterable

from ssort._ast import iter_child_nodes
from ssort._utils import single_dispatch


@single_dispatch
def _get_attribute_accesses(node: ast.AST, variable: str) -> Iterable[str]:
    for child in iter_child_nodes(node):
        yield from _get_attribute_accesses(child, variable)


@_get_attribute_accesses.register(ast.ClassDef)
def _get_attribute_accesses_for_class_def(
    node: ast.ClassDef, variable: str
) -> Iterable[str]:
    # TODO
    return ()


@_get_attribute_accesses.register(ast.Attribute)
def _get_attribute_accesses_for_attribute(
    node: ast.Attribute, variable: str
) -> Iterable[str]:
    yield from _get_attribute_accesses(node.value, variable)
    if (
        isinstance(node.ctx, ast.Load)
        and isinstance(node.value, ast.Name)
        and node.value.id == variable
    ):
        yield node.attr


@single_dispatch
def get_method_requirements(node: ast.AST) -> Iterable[str]:
    return ()


@get_method_requirements.register(ast.FunctionDef)
@get_method_requirements.register(ast.AsyncFunctionDef)
def _get_method_requirements_for_function_def(
    node: ast.FunctionDef | ast.AsyncFunctionDef,
) -> Iterable[str]:
    if node.args.posonlyargs:
        self_arg = node.args.posonlyargs[0].arg
    elif node.args.args:
        self_arg = node.args.args[0].arg
    else:
        return

    for statement in node.body:
        yield from _get_attribute_accesses(statement, self_arg)



import argparse
import difflib
import re
import sys

from ssort import __version__
from ssort._exceptions import UnknownEncodingError
from ssort._files import find_python_files
from ssort._ssort import ssort
from ssort._utils import (
    detect_encoding,
    detect_newline,
    escape_path,
    normalize_newlines,
)


def main():
    parser = argparse.ArgumentParser(
        description="Sort python statements into dependency order",
    )

    parser.add_argument(
        "--version",
        dest="version",
        action="store_true",
        help="Outputs version information and then exit",
    )
    parser.add_argument(
        "--diff",
        dest="show_diff",
        action="store_true",
        help="Prints a diff of all changes ssort would make to a file.",
    )
    parser.add_argument(
        "--check",
        dest="check",
        action="store_true",
        help="Check the file for unsorted statements.  Returns 0 if nothing "
        "needs to be changed.  Otherwise returns 1.",
    )
    parser.add_argument(
        "files", nargs="*", help="One or more python files to sort"
    )

    args = parser.parse_args()

    if args.version:
        sys.stdout.write(f"ssort {__version__}\n")
        return

    unsorted = 0
    unsortable = 0
    unchanged = 0

    for path in find_python_files(args.files):
        errors = False

        try:
            original_bytes = path.read_bytes()
        except FileNotFoundError:
            sys.stderr.write(f"ERROR: {escape_path(path)} does not exist\n")
            unsortable += 1
            continue
        except IsADirectoryError:
            sys.stderr.write(f"ERROR: {escape_path(path)} is a directory\n")
            unsortable += 1
            continue
        except PermissionError:
            sys.stderr.write(f"ERROR: {escape_path(path)} is not readable\n")
            unsortable += 1
            continue

        # The logic for converting from bytes to text is duplicated in `ssort`
        # and here because we need access to the text to be able to compute a
        # diff at the end.
        try:
            encoding = detect_encoding(original_bytes)
        except UnknownEncodingError as exc:
            sys.stderr.write(
                f"ERROR: unknown encoding, {exc.encoding!r}, in {escape_path(path)}\n"
            )
            unsortable += 1
            continue

        try:
            original = original_bytes.decode(encoding)
        except UnicodeDecodeError as exc:
            sys.stderr.write(
                f"ERROR: encoding error in {escape_path(path)}: {exc}\n"
            )
            unsortable += 1
            continue

        newline = detect_newline(original)
        original = normalize_newlines(original)

        def _on_parse_error(message, *, lineno, col_offset, **kwargs):
            nonlocal errors
            errors = True

            sys.stderr.write(
                f"ERROR: syntax error in {escape_path(path)}: "
                + f"line {lineno}, column {col_offset}\n"
            )

        def _on_unresolved(message, *, name, lineno, col_offset, **kwargs):
            nonlocal errors
            errors = True

            sys.stderr.write(
                f"ERROR: unresolved dependency {name!r} "
                + f"in {escape_path(path)}: "
                + f"line {lineno}, column {col_offset}\n"
            )

        def _on_wildcard_import(**kwargs):
            sys.stderr.write(
                "WARNING: can't determine dependencies on * import\n"
            )

        try:
            updated = ssort(
                original,
                filename=escape_path(path),
                on_parse_error=_on_parse_error,
                on_unresolved=_on_unresolved,
                on_wildcard_import=_on_wildcard_import,
            )

            if errors:
                unsortable += 1
                continue

        except Exception as e:
            raise Exception(f"ERROR while sorting {path}\n") from e

        if original != updated:
            unsorted += 1
            if args.check:
                sys.stderr.write(
                    f"ERROR: {escape_path(path)} is incorrectly sorted\n"
                )
            else:
                sys.stderr.write(f"Sorting {escape_path(path)}\n")

                # The logic for converting from bytes to text is duplicated in
                # `ssort` and here because we need access to the text to be able
                # to compute a diff at the end.
                # We rename a little prematurely to avoid shadowing `updated`,
                # which we use later for printing the diff.
                updated_bytes = updated
                if newline != "\n":
                    updated_bytes = re.sub("\n", newline, updated_bytes)
                updated_bytes = updated_bytes.encode(encoding)

                path.write_bytes(updated_bytes)
        else:
            unchanged += 1

        if args.show_diff:
            sys.stderr.writelines(
                difflib.unified_diff(
                    original.splitlines(keepends=True),
                    updated.splitlines(keepends=True),
                    fromfile=f"{path}:before",
                    tofile=f"{path}:after",
                )
            )

    if args.check:

        def _fmt_count(count):
            return f"{count} file" if count == 1 else f"{count} files"

        summary = []
        if unsorted:
            summary.append(f"{_fmt_count(unsorted)} would be resorted")
        if unchanged:
            summary.append(f"{_fmt_count(unchanged)} would be left unchanged")
        if unsortable:
            summary.append(f"{_fmt_count(unsortable)} would not be sortable")
        if not unsorted and not unchanged and not unsortable:
            summary.append("No files are present to be sorted. Nothing to do.")

        sys.stderr.write(", ".join(summary) + "\n")

        if unsorted or unsortable:
            sys.exit(1)

    else:

        def _fmt_count_were(count):
            if count == 1:
                return f"{count} file was"
            else:
                return f"{count} files were"

        summary = []
        if unsorted:
            summary.append(f"{_fmt_count_were(unsorted)} resorted")
        if unchanged:
            summary.append(f"{_fmt_count_were(unchanged)} left unchanged")
        if unsortable:
            summary.append(f"{_fmt_count_were(unsortable)} not sortable")
        if not unsorted and not unchanged and not unsortable:
            summary.append("No files are present to be sorted. Nothing to do.")

        sys.stderr.write(", ".join(summary) + "\n")

        if unsortable:
            sys.exit(1)



import ast
import warnings
from io import StringIO
from token import NAME
from tokenize import generate_tokens

from ssort._exceptions import ParseError
from ssort._statements import Statement


def _find_start(node):
    if (
        isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef))
        and node.decorator_list
    ):
        decorator = node.decorator_list[0]
        return decorator.lineno - 1, decorator.col_offset
    else:
        return node.lineno - 1, node.col_offset


def _find_end(node):
    return node.end_lineno - 1, node.end_col_offset


def split(
    root_text,
    *,
    nodes,
    next_row=0,
    next_col=0,
    indent=0,
):
    # Build an index of row lengths and start offsets to enable fast string
    # indexing using ast row/column coordinates.
    row_lengths = []
    row_offsets = [0]
    for offset, char in enumerate(root_text):
        if char == "\n":
            row_lengths.append(offset - row_offsets[-1])
            row_offsets.append(offset + 1)
    row_lengths.append(len(root_text) - row_offsets[-1])

    nodes = iter(nodes)

    next_node = next(nodes, None)

    if next_node is not None:
        next_start_row, next_start_col = _find_start(next_node)
        next_end_row, next_end_col = _find_end(next_node)

        indent_text = " " * next_node.col_offset
        next_indent_text = ""

    while next_node:
        this_node, next_node = next_node, next(nodes, None)
        this_end_row, this_end_col = next_end_row, next_end_col
        this_indent_text = next_indent_text

        if next_node is not None:
            next_start_row, next_start_col = _find_start(next_node)
            next_end_row, next_end_col = _find_end(next_node)

        start_row = next_row
        start_col = next_col

        if next_node is not None and this_end_row == next_end_row:
            # There is another statement on the same line.  It should be
            # possible to claim as far as the start of the next node for this
            # node, but this space can only contain semicolons and whitespace
            # so we are better off filtering it out.
            end_row = this_end_row
            end_col = this_end_col

            next_row = next_start_row
            next_col = next_start_col

            next_indent_text = indent_text
        else:
            # No other statements on the same line.  Assume that everything up
            # until the end of the line is comments attached to this statement.
            end_row = this_end_row
            end_col = row_lengths[end_row]

            next_row = this_end_row + 1
            next_col = 0

            next_indent_text = ""

        start_offset = row_offsets[start_row] + start_col
        if next_node is not None:
            end_offset = row_offsets[end_row] + end_col
        else:
            end_offset = len(root_text.rstrip("\n"))

        yield Statement(
            text=this_indent_text + root_text[start_offset:end_offset],
            node=this_node,
            start_row=start_row,
            start_col=start_col,
        )


def split_class(statement):
    node = statement.node
    text = statement.text
    text_padded = statement.text_padded()

    # Build an index of row lengths and start offsets to enable fast string
    # indexing using ast row/column coordinates.
    row_lengths = []
    row_offsets = [0]
    for offset, char in enumerate(text_padded):
        if char == "\n":
            row_lengths.append(offset - row_offsets[-1])
            row_offsets.append(offset + 1)
    row_lengths.append(len(text_padded) - row_offsets[-1])

    tokens = iter(generate_tokens(StringIO(text_padded).readline))

    for token in tokens:
        lineno, col_offset = token.start
        if lineno == node.lineno and col_offset == node.col_offset:
            assert token.string == "class"
            break

    token = next(tokens)
    assert token.type == NAME

    token = next(tokens)
    if token.string == "(":
        token = next(tokens)
        depth = 1
        while depth:
            if token.string == "(":
                depth += 1
            if token.string == ")":
                depth -= 1
            token = next(tokens)

    assert token.string == ":"

    if node.body[0].lineno == token.end[0]:
        # All tokens are on the same line.  `split` won't know how to indent
        # them so we do it ourselves.
        head_end_lineno, head_end_col = token.end
        head_end_row = head_end_lineno - 1

        head_end_offset = row_offsets[head_end_row] + head_end_col
        head_text_padded = text_padded[:head_end_offset].rstrip()
        head_text = head_text_padded[len(text_padded) - len(text) :]

        body_statements = []
        for child_node in node.body:
            child_start_row, child_start_col = _find_start(child_node)
            child_end_row, child_end_col = _find_end(child_node)

            assert child_start_row == head_end_row
            assert child_end_row == head_end_row

            start_offset = row_offsets[child_start_row] + child_start_col
            end_offset = row_offsets[child_end_row] + child_end_col

            body_statements.append(
                Statement(
                    text="    " + text_padded[start_offset:end_offset],
                    node=child_node,
                    start_row=child_start_row,
                    start_col=child_start_col,
                )
            )

    else:
        head_end_lineno, head_end_col = token.end[0] + 1, 0
        head_end_row = head_end_lineno - 1

        head_end_offset = row_offsets[head_end_row] + head_end_col
        head_text_padded = text_padded[:head_end_offset].rstrip()
        head_text = head_text_padded[len(text_padded) - len(text) :]

        body_text_padded = (
            (head_end_row) * "\n"
            + head_end_col * " "
            + text_padded[head_end_offset:]
        )

        body_statements = list(
            split(
                body_text_padded,
                nodes=node.body,
                next_row=head_end_row,
                next_col=head_end_col,
            )
        )

    return head_text, body_statements


def parse(root_text, *, filename="<unknown>"):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            root_node = ast.parse(root_text, filename)
        except SyntaxError as exc:
            raise ParseError(exc.msg, lineno=exc.lineno, col_offset=exc.offset)
    return split(root_text, nodes=list(root_node.body))



from __future__ import annotations

from typing import Callable, Generic, Hashable, TypeVar

from ssort._utils import sort_key_from_iter

_T = TypeVar("_T", bound=Hashable)


class Graph(Generic[_T]):
    def __init__(self) -> None:
        self.nodes: list[_T] = []
        self.dependencies: dict[_T, list[_T]] = {}
        self.dependants: dict[_T, list[_T]] = {}

    def add_node(self, identifier: _T) -> None:
        if identifier not in self.nodes:
            self.nodes.append(identifier)
            self.dependencies[identifier] = []
            self.dependants[identifier] = []

    def add_dependency(self, node: _T, dependency: _T) -> None:
        assert dependency in self.nodes

        if dependency not in self.dependencies[node]:
            self.dependencies[node].append(dependency)
            self.dependants[dependency].append(node)

    def remove_node(self, node: _T) -> None:
        self.nodes.remove(node)
        del self.dependencies[node]
        del self.dependants[node]

        for other in self.nodes:
            try:
                self.dependencies[other].remove(node)
            except ValueError:
                pass

            try:
                self.dependants[other].remove(node)
            except ValueError:
                pass

    def remove_dependency(self, node: _T, dependency: _T) -> None:
        assert dependency in self.nodes

        try:
            self.dependencies[node].remove(dependency)
        except ValueError:
            pass

        try:
            self.dependants[dependency].remove(node)
        except ValueError:
            pass

    def update(self, other: Graph[_T]) -> None:
        for node in other.nodes:
            self.add_node(node)

        for node in other.nodes:
            for dependency in other.dependencies[node]:
                self.add_dependency(node, dependency)

    def copy(self) -> Graph[_T]:
        dup: Graph[_T] = Graph()
        dup.update(self)
        return dup


def _remove_self_references(graph: Graph[_T]) -> None:
    for node in graph.nodes:
        graph.remove_dependency(node, node)


def _find_cycle(graph: Graph[_T]) -> list[_T] | None:
    processed = set()
    for node in graph.nodes:
        if node in processed:
            continue

        in_stack = {node}
        stack = [(node, set(graph.dependencies[node]))]

        while stack:
            top_node, top_dependencies = stack[-1]

            if not top_dependencies:
                processed.add(top_node)
                in_stack.remove(top_node)
                stack.pop()
                continue

            dependency = top_dependencies.pop()
            if dependency in in_stack:
                cycle = [dependency]
                while stack[-1][0] != dependency:
                    cycle.append(stack[-1][0])
                    stack.pop()
                return cycle
            if dependency not in processed:
                stack.append((dependency, set(graph.dependencies[dependency])))
                in_stack.add(dependency)

    return None


def replace_cycles(graph: Graph[_T], *, key: Callable[[_T], int]) -> None:
    """
    Finds all cycles and replaces them with forward links that keep them from
    being re-ordered.
    """
    _remove_self_references(graph)
    while True:
        cycle = _find_cycle(graph)
        if not cycle:
            break

        for node in cycle:
            for dependency in cycle:
                graph.remove_dependency(node, dependency)

        # TODO this is a bit of an abstraction leak.  Need a better way to tell
        # this function what the safe order is.
        nodes = iter(sorted(cycle, key=key))
        prev = next(nodes)
        for node in nodes:
            graph.add_dependency(node, prev)
            prev = node


def is_topologically_sorted(nodes: list[_T], graph: Graph[_T]) -> bool:
    visited = set()
    for node in nodes:
        visited.add(node)
        for dependency in graph.dependencies[node]:
            if dependency not in visited:
                return False
    return True


def topological_sort(
    target: Graph[_T] | list[_T], /, *, graph: Graph[_T] | None = None
) -> list[_T]:
    if graph is None:
        if not isinstance(target, Graph):
            raise TypeError("target must be a Graph")
        graph = target
        nodes = target.nodes
    else:
        if not isinstance(target, list):
            raise TypeError("target must be a list")
        nodes = target

    # Create a mutable copy of the graph so that we can pop edges from it as we
    # traverse.
    remaining = graph.copy()

    key = sort_key_from_iter(nodes)

    pending = [node for node in graph.nodes if not graph.dependants[node]]

    result = []
    while pending:
        pending = list(sorted(pending, key=key))
        node = pending.pop()
        dependencies = remaining.dependencies[node]
        remaining.remove_node(node)

        for dependency in dependencies:
            if not remaining.dependants[dependency]:
                if dependency not in pending:
                    pending.append(dependency)

        result.append(node)

    result.reverse()

    assert not remaining.nodes
    assert is_topologically_sorted(result, graph)

    return [node for node in result if node in nodes]



from __future__ import annotations

import ast
import sys
from typing import Iterable

from ssort._utils import single_dispatch


@single_dispatch
def iter_child_nodes(node: ast.AST) -> Iterable[ast.AST]:
    raise NotImplementedError(
        f"AST traversal for {type(node).__name__!r} is not implemented"
    )


@iter_child_nodes.register(ast.Module)
def _iter_child_nodes_of_module(node: ast.Module) -> Iterable[ast.AST]:
    yield from node.body
    yield from node.type_ignores


@iter_child_nodes.register(ast.Interactive)
def _iter_child_nodes_of_interactive(
    node: ast.Interactive,
) -> Iterable[ast.AST]:
    yield from node.body


@iter_child_nodes.register(ast.Expression)
def _iter_child_nodes_of_expression(node: ast.Expression) -> Iterable[ast.AST]:
    yield node.body


@iter_child_nodes.register(ast.FunctionType)
def _iter_child_nodes_of_function_type(
    node: ast.FunctionType,
) -> Iterable[ast.AST]:
    yield from node.argtypes
    yield node.returns


@iter_child_nodes.register(ast.FunctionDef)
@iter_child_nodes.register(ast.AsyncFunctionDef)
def _iter_child_nodes_of_function_def(
    node: ast.FunctionDef | ast.AsyncFunctionDef,
) -> Iterable[ast.AST]:
    yield from node.decorator_list
    yield node.args
    if node.returns is not None:
        yield node.returns
    yield from node.body


@iter_child_nodes.register(ast.ClassDef)
def _iter_child_nodes_of_class_def(node: ast.ClassDef) -> Iterable[ast.AST]:
    yield from node.decorator_list
    yield from node.bases
    yield from node.keywords
    yield from node.body


@iter_child_nodes.register(ast.Return)
def _iter_child_nodes_of_return(node: ast.Return) -> Iterable[ast.AST]:
    if node.value is not None:
        yield node.value


@iter_child_nodes.register(ast.Delete)
def _iter_child_nodes_of_delete(node: ast.Delete) -> Iterable[ast.AST]:
    yield from node.targets


@iter_child_nodes.register(ast.Assign)
def _iter_child_nodes_of_assign(node: ast.Assign) -> Iterable[ast.AST]:
    yield from node.targets
    yield node.value


@iter_child_nodes.register(ast.AugAssign)
def _iter_child_nodes_of_aug_assign(node: ast.AugAssign) -> Iterable[ast.AST]:
    yield node.target
    yield node.value


@iter_child_nodes.register(ast.AnnAssign)
def _iter_child_nodes_of_ann_assign(node: ast.AnnAssign) -> Iterable[ast.AST]:
    yield node.target
    yield node.annotation
    if node.value is not None:
        yield node.value


@iter_child_nodes.register(ast.For)
@iter_child_nodes.register(ast.AsyncFor)
def _iter_child_nodes_of_for(
    node: ast.For | ast.AsyncFor,
) -> Iterable[ast.AST]:
    yield node.target
    yield node.iter
    yield from node.body
    yield from node.orelse


@iter_child_nodes.register(ast.While)
def _iter_child_nodes_of_while(node: ast.While) -> Iterable[ast.AST]:
    yield node.test
    yield from node.body
    yield from node.orelse


@iter_child_nodes.register(ast.If)
def _iter_child_nodes_of_if(node: ast.If) -> Iterable[ast.AST]:
    yield node.test
    yield from node.body
    yield from node.orelse


@iter_child_nodes.register(ast.With)
@iter_child_nodes.register(ast.AsyncWith)
def _iter_child_nodes_of_with(
    node: ast.With | ast.AsyncWith,
) -> Iterable[ast.AST]:
    yield from node.items
    yield from node.body


if sys.version_info >= (3, 10):

    @iter_child_nodes.register(ast.Match)
    def _iter_child_nodes_of_match(node: ast.Match) -> Iterable[ast.AST]:
        yield node.subject
        yield from node.cases


@iter_child_nodes.register(ast.Raise)
def _iter_child_nodes_of_raise(node: ast.Raise) -> Iterable[ast.AST]:
    if node.exc is not None:
        yield node.exc
    if node.cause is not None:
        yield node.cause


@iter_child_nodes.register(ast.Try)
def _iter_child_nodes_of_try(node: ast.Try) -> Iterable[ast.AST]:
    yield from node.body
    yield from node.handlers
    yield from node.orelse
    yield from node.finalbody


@iter_child_nodes.register(ast.Assert)
def _iter_child_nodes_of_assert(node: ast.Assert) -> Iterable[ast.AST]:
    yield node.test
    if node.msg is not None:
        yield node.msg


@iter_child_nodes.register(ast.Import)
def _iter_child_nodes_of_import(node: ast.Import) -> Iterable[ast.AST]:
    yield from node.names


@iter_child_nodes.register(ast.ImportFrom)
def _iter_child_nodes_of_import_from(
    node: ast.ImportFrom,
) -> Iterable[ast.AST]:
    yield from node.names


@iter_child_nodes.register(ast.Global)
@iter_child_nodes.register(ast.Nonlocal)
def _iter_child_nodes_of_scope(
    node: ast.Global | ast.Nonlocal,
) -> Iterable[ast.AST]:
    return ()


@iter_child_nodes.register(ast.Expr)
def _iter_child_nodes_of_expr(node: ast.Expr) -> Iterable[ast.AST]:
    yield node.value


@iter_child_nodes.register(ast.Pass)
@iter_child_nodes.register(ast.Break)
@iter_child_nodes.register(ast.Continue)
def _iter_child_nodes_of_control_flow(
    node: ast.Pass | ast.Break | ast.Continue,
) -> Iterable[ast.AST]:
    return ()


@iter_child_nodes.register(ast.BoolOp)
def _iter_child_nodes_of_bool_op(node: ast.BoolOp) -> Iterable[ast.AST]:
    yield from node.values


@iter_child_nodes.register(ast.NamedExpr)
def _iter_child_nodes_of_named_expr(node: ast.NamedExpr) -> Iterable[ast.AST]:
    yield node.target
    yield node.value


@iter_child_nodes.register(ast.BinOp)
def _iter_child_nodes_of_bin_op(node: ast.BinOp) -> Iterable[ast.AST]:
    yield node.left
    yield node.right


@iter_child_nodes.register(ast.UnaryOp)
def _iter_child_nodes_of_unary_op(node: ast.UnaryOp) -> Iterable[ast.AST]:
    yield node.operand


@iter_child_nodes.register(ast.Lambda)
def _iter_child_nodes_of_lambda(node: ast.Lambda) -> Iterable[ast.AST]:
    yield node.args
    yield node.body


@iter_child_nodes.register(ast.IfExp)
def _iter_child_nodes_of_if_exp(node: ast.IfExp) -> Iterable[ast.AST]:
    yield node.test
    yield node.body
    yield node.orelse


@iter_child_nodes.register(ast.Dict)
def _iter_child_nodes_of_dict(node: ast.Dict) -> Iterable[ast.AST]:
    for key in node.keys:
        if key is not None:
            yield key
    yield from node.values


@iter_child_nodes.register(ast.Set)
def _iter_child_nodes_of_set(node: ast.Set) -> Iterable[ast.AST]:
    yield from node.elts


@iter_child_nodes.register(ast.ListComp)
def _iter_child_nodes_of_list_comp(node: ast.ListComp) -> Iterable[ast.AST]:
    yield node.elt
    yield from node.generators


@iter_child_nodes.register(ast.SetComp)
def _iter_child_nodes_of_set_comp(node: ast.SetComp) -> Iterable[ast.AST]:
    yield node.elt
    yield from node.generators


@iter_child_nodes.register(ast.DictComp)
def _iter_child_nodes_of_dict_comp(node: ast.DictComp) -> Iterable[ast.AST]:
    yield node.key
    yield node.value
    yield from node.generators


@iter_child_nodes.register(ast.GeneratorExp)
def _iter_child_nodes_of_generator_exp(
    node: ast.GeneratorExp,
) -> Iterable[ast.AST]:
    yield node.elt
    yield from node.generators


@iter_child_nodes.register(ast.Await)
def _iter_child_nodes_of_await(node: ast.Await) -> Iterable[ast.AST]:
    yield node.value


@iter_child_nodes.register(ast.Yield)
def _iter_child_nodes_of_yield(node: ast.Yield) -> Iterable[ast.AST]:
    if node.value is not None:
        yield node.value


@iter_child_nodes.register(ast.YieldFrom)
def _iter_child_nodes_of_yield_from(node: ast.YieldFrom) -> Iterable[ast.AST]:
    yield node.value


@iter_child_nodes.register(ast.Compare)
def _iter_child_nodes_of_compare(node: ast.Compare) -> Iterable[ast.AST]:
    yield node.left
    yield from node.comparators


@iter_child_nodes.register(ast.Call)
def _iter_child_nodes_of_call(node: ast.Call) -> Iterable[ast.AST]:
    yield node.func
    yield from node.args
    yield from node.keywords


@iter_child_nodes.register(ast.FormattedValue)
def _iter_child_nodes_of_formatted_value(
    node: ast.FormattedValue,
) -> Iterable[ast.AST]:
    yield node.value
    if node.format_spec is not None:
        yield node.format_spec


@iter_child_nodes.register(ast.JoinedStr)
def _iter_child_nodes_of_joined_str(node: ast.JoinedStr) -> Iterable[ast.AST]:
    yield from node.values


@iter_child_nodes.register(ast.Constant)
def _iter_child_nodes_of_constant(node: ast.Constant) -> Iterable[ast.AST]:
    return ()


@iter_child_nodes.register(ast.Attribute)
def _iter_child_nodes_of_attribute(node: ast.Attribute) -> Iterable[ast.AST]:
    yield node.value


@iter_child_nodes.register(ast.Subscript)
def _iter_child_nodes_of_subscript(node: ast.Subscript) -> Iterable[ast.AST]:
    yield node.value
    yield node.slice


@iter_child_nodes.register(ast.Starred)
def _iter_child_nodes_of_starred(node: ast.Starred) -> Iterable[ast.AST]:
    yield node.value


@iter_child_nodes.register(ast.Name)
def _iter_child_nodes_of_name(node: ast.Name) -> Iterable[ast.AST]:
    return ()


@iter_child_nodes.register(ast.List)
@iter_child_nodes.register(ast.Tuple)
def _iter_child_nodes_of_sequence(
    node: ast.List | ast.Tuple,
) -> Iterable[ast.AST]:
    yield from node.elts


@iter_child_nodes.register(ast.Slice)
def _iter_child_nodes_of_slice(node: ast.Slice) -> Iterable[ast.AST]:
    if node.lower is not None:
        yield node.lower
    if node.upper is not None:
        yield node.upper
    if node.step is not None:
        yield node.step


if sys.version_info < (3, 9):

    @iter_child_nodes.register(ast.ExtSlice)
    def _iter_child_nodes_of_ext_slice(
        node: ast.ExtSlice,
    ) -> Iterable[ast.AST]:
        yield from node.dims

    @iter_child_nodes.register(ast.Index)
    def _iter_child_nodes_of_index(node: ast.Index) -> Iterable[ast.AST]:
        yield node.value


@iter_child_nodes.register(ast.comprehension)
def _iter_child_nodes_of_comprehension(
    node: ast.comprehension,
) -> Iterable[ast.AST]:
    yield node.target
    yield node.iter
    yield from node.ifs


@iter_child_nodes.register(ast.ExceptHandler)
def _iter_child_nodes_of_except_handler(
    node: ast.ExceptHandler,
) -> Iterable[ast.AST]:
    if node.type is not None:
        yield node.type
    yield from node.body


@iter_child_nodes.register(ast.arguments)
def _iter_child_nodes_of_arguments(node: ast.arguments) -> Iterable[ast.AST]:
    yield from node.posonlyargs
    yield from node.args
    if node.vararg is not None:
        yield node.vararg
    yield from node.kwonlyargs
    for default in node.kw_defaults:
        if default is not None:
            yield default
    if node.kwarg is not None:
        yield node.kwarg
    yield from node.defaults


@iter_child_nodes.register(ast.arg)
def _iter_child_nodes_of_arg(node: ast.arg) -> Iterable[ast.AST]:
    if node.annotation is not None:
        yield node.annotation


@iter_child_nodes.register(ast.keyword)
def _iter_child_nodes_of_keyword(node: ast.keyword) -> Iterable[ast.AST]:
    yield node.value


@iter_child_nodes.register(ast.alias)
def _iter_child_nodes_of_alias(node: ast.alias) -> Iterable[ast.AST]:
    return ()


@iter_child_nodes.register(ast.withitem)
def _iter_child_nodes_of_withitem(node: ast.withitem) -> Iterable[ast.AST]:
    yield node.context_expr
    if node.optional_vars is not None:
        yield node.optional_vars


if sys.version_info >= (3, 10):

    @iter_child_nodes.register(ast.match_case)
    def _iter_child_nodes_of_match_case(
        node: ast.match_case,
    ) -> Iterable[ast.AST]:
        yield node.pattern
        if node.guard is not None:
            yield node.guard
        yield from node.body

    @iter_child_nodes.register(ast.MatchValue)
    def _iter_child_nodes_of_match_value(
        node: ast.MatchValue,
    ) -> Iterable[ast.AST]:
        yield node.value

    @iter_child_nodes.register(ast.MatchSingleton)
    def _iter_child_nodes_of_match_singleton(
        node: ast.MatchSingleton,
    ) -> Iterable[ast.AST]:
        return ()

    @iter_child_nodes.register(ast.MatchSequence)
    def _iter_child_nodes_of_match_sequence(
        node: ast.MatchSequence,
    ) -> Iterable[ast.AST]:
        yield from node.patterns

    @iter_child_nodes.register(ast.MatchMapping)
    def _iter_child_nodes_of_match_mapping(
        node: ast.MatchMapping,
    ) -> Iterable[ast.AST]:
        yield from node.keys
        yield from node.patterns

    @iter_child_nodes.register(ast.MatchClass)
    def _iter_child_nodes_of_match_class(
        node: ast.MatchClass,
    ) -> Iterable[ast.AST]:
        yield node.cls
        yield from node.patterns
        yield from node.kwd_patterns

    @iter_child_nodes.register(ast.MatchStar)
    def _iter_child_nodes_of_match_star(
        node: ast.MatchStar,
    ) -> Iterable[ast.AST]:
        return ()

    @iter_child_nodes.register(ast.MatchAs)
    def _iter_child_nodes_of_match_as(node: ast.MatchAs) -> Iterable[ast.AST]:
        if node.pattern is not None:
            yield node.pattern

    @iter_child_nodes.register(ast.MatchOr)
    def _iter_child_nodes_of_match_or(node: ast.MatchOr) -> Iterable[ast.AST]:
        yield from node.patterns


if sys.version_info >= (3, 11):

    @iter_child_nodes.register(ast.TryStar)
    def _iter_child_nodes_of_trystar(node: ast.TryStar) -> Iterable[ast.AST]:
        yield from node.body
        yield from node.handlers
        yield from node.orelse
        yield from node.finalbody


@iter_child_nodes.register(ast.TypeIgnore)
def _iter_child_nodes_of_type_ignore(
    node: ast.TypeIgnore,
) -> Iterable[ast.AST]:
    return ()



from __future__ import annotations

import ast
import dataclasses
import enum
from typing import Iterable

from ssort._ast import iter_child_nodes
from ssort._bindings import get_bindings
from ssort._builtins import CLASS_BUILTINS
from ssort._utils import single_dispatch


class Scope(enum.Enum):
    LOCAL = "LOCAL"
    NONLOCAL = "NONLOCAL"
    GLOBAL = "GLOBAL"


@dataclasses.dataclass(frozen=True)
class Requirement:
    name: str
    lineno: int
    col_offset: int
    deferred: bool = False
    scope: Scope = Scope.LOCAL


@single_dispatch
def get_requirements(node: ast.AST) -> Iterable[Requirement]:
    for child in iter_child_nodes(node):
        yield from get_requirements(child)


def _get_scope_from_arguments(args: ast.arguments) -> set[str]:
    scope: set[str] = set()
    scope.update(arg.arg for arg in args.posonlyargs)
    scope.update(arg.arg for arg in args.args)  # Arghhh.
    if args.vararg:
        scope.add(args.vararg.arg)
    scope.update(arg.arg for arg in args.kwonlyargs)
    if args.kwarg:
        scope.add(args.kwarg.arg)
    return scope


@get_requirements.register(ast.FunctionDef)
@get_requirements.register(ast.AsyncFunctionDef)
def _get_requirements_for_function_def(
    node: ast.FunctionDef | ast.AsyncFunctionDef,
) -> Iterable[Requirement]:
    for decorator in node.decorator_list:
        yield from get_requirements(decorator)

    yield from get_requirements(node.args)

    if node.returns is not None:
        yield from get_requirements(node.returns)

    scope = _get_scope_from_arguments(node.args)

    requirements = []
    for statement in node.body:
        scope.update(get_bindings(statement))
        for requirement in get_requirements(statement):
            if not requirement.deferred:
                requirement = dataclasses.replace(requirement, deferred=True)
            requirements.append(requirement)

    for requirement in requirements:
        if requirement.scope == Scope.GLOBAL:
            yield requirement
        elif requirement.scope == Scope.NONLOCAL:
            yield dataclasses.replace(requirement, scope=Scope.LOCAL)
        elif requirement.name not in scope:
            yield requirement


@get_requirements.register(ast.ClassDef)
def _get_requirements_for_class_def(
    node: ast.ClassDef,
) -> Iterable[Requirement]:
    for decorator in node.decorator_list:
        yield from get_requirements(decorator)

    for base in node.bases:
        yield from get_requirements(base)

    scope = set(CLASS_BUILTINS)

    for statement in node.body:
        for stmt_dep in get_requirements(statement):
            if stmt_dep.deferred or stmt_dep.name not in scope:
                yield stmt_dep

        scope.update(get_bindings(statement))


@get_requirements.register(ast.For)
@get_requirements.register(ast.AsyncFor)
def _get_requirements_for_for(
    node: ast.For | ast.AsyncFor,
) -> Iterable[Requirement]:
    bindings = set(get_bindings(node))

    yield from get_requirements(node.target)
    yield from get_requirements(node.iter)

    for stmt in node.body:
        for requirement in get_requirements(stmt):
            if requirement.name not in bindings:
                yield requirement

    for stmt in node.orelse:
        for requirement in get_requirements(stmt):
            if requirement.name not in bindings:
                yield requirement


@get_requirements.register(ast.With)
@get_requirements.register(ast.AsyncWith)
def _get_requirements_for_with(
    node: ast.With | ast.AsyncWith,
) -> Iterable[Requirement]:
    bindings = set(get_bindings(node))

    for item in node.items:
        yield from get_requirements(item)

    for stmt in node.body:
        for requirement in get_requirements(stmt):
            if requirement.name not in bindings:
                yield requirement


@get_requirements.register(ast.Global)
def _get_requirements_for_global(node: ast.Global) -> Iterable[Requirement]:
    for name in node.names:
        yield Requirement(
            name=name,
            lineno=node.lineno,
            col_offset=node.col_offset,
            scope=Scope.GLOBAL,
        )


@get_requirements.register(ast.Nonlocal)
def _get_requirements_for_nonlocal(
    node: ast.Nonlocal,
) -> Iterable[Requirement]:
    for name in node.names:
        yield Requirement(
            name=name,
            lineno=node.lineno,
            col_offset=node.col_offset,
            scope=Scope.NONLOCAL,
        )


@get_requirements.register(ast.Lambda)
def _get_requirements_for_lambda(node: ast.Lambda) -> Iterable[Requirement]:
    yield from get_requirements(node.args)
    scope = _get_scope_from_arguments(node.args)
    scope.update(get_bindings(node.body))
    for requirement in get_requirements(node.body):
        if requirement.name not in scope:
            yield requirement


@get_requirements.register(ast.ListComp)
@get_requirements.register(ast.SetComp)
@get_requirements.register(ast.DictComp)
@get_requirements.register(ast.GeneratorExp)
def _get_requirements_for_comp(
    node: ast.ListComp | ast.SetComp | ast.DictComp | ast.GeneratorExp,
) -> Iterable[Requirement]:
    bindings = {
        binding
        for generator in node.generators
        for binding in get_bindings(generator.target)
    }
    for child in iter_child_nodes(node):
        for requirement in get_requirements(child):
            if requirement.name not in bindings:
                yield requirement


@get_requirements.register(ast.Name)
def _get_requirements_for_name(node: ast.Name) -> Iterable[Requirement]:
    if isinstance(node.ctx, (ast.Load, ast.Del)):
        yield Requirement(
            name=node.id, lineno=node.lineno, col_offset=node.col_offset
        )



from __future__ import annotations

import os
import pathlib
from typing import Iterable

import pathspec

from ssort._utils import memoize

_EMPTY_PATH_SPEC = pathspec.PathSpec([])


@memoize
def _is_project_root(path: pathlib.Path) -> bool:
    if path == path.root or path == path.parent:
        return True

    if (path / ".git").is_dir():
        return True

    return False


@memoize
def _get_ignore_patterns(path: pathlib.Path) -> pathspec.PathSpec:
    git_ignore = path / ".gitignore"
    if git_ignore.is_file():
        with git_ignore.open() as f:
            return pathspec.PathSpec.from_lines("gitwildmatch", f)

    return _EMPTY_PATH_SPEC


def is_ignored(path: str | os.PathLike) -> bool:
    # Can't use pathlib.Path.resolve() here because we want to maintain
    # symbolic links.
    path = pathlib.Path(os.path.abspath(path))

    for part in (path, *path.parents):
        patterns = _get_ignore_patterns(part)
        if patterns.match_file(path.relative_to(part)):
            return True

        if _is_project_root(part):
            return False

    return False


def find_python_files(
    patterns: Iterable[str | os.PathLike[str]],
) -> Iterable[pathlib.Path]:
    if not patterns:
        patterns = ["."]

    paths_set = set()
    for pattern in patterns:
        path = pathlib.Path(pattern)
        if not path.is_dir():
            subpaths = [path]
        else:
            subpaths = [
                subpath
                for subpath in path.glob("**/*.py")
                if not is_ignored(subpath) and subpath.is_file()
            ]

        for subpath in sorted(subpaths):
            if subpath not in paths_set:
                paths_set.add(subpath)
                yield subpath
